#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <norm.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <net/ethernet.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>

#ifdef FREEBSD
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#endif
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>
#include <pcap.h>

#ifdef LINUX
struct ipovly
{
  u_char           ih_x1[9];
  u_char           ih_pr;
  u_int16_t        ih_len;
  struct in_addr   ih_src;
  struct in_addr   ih_dst;
};
#endif


#define FDDI_HDR_LEN    13
#define SNAP_LEN        8

static char             errbuf[PCAP_ERRBUF_SIZE];
static pcap_t         *pcap;
static char           *input;
static char           *output;
static pcap_dumper_t  *dumper;   

static long int        acc_size    = 0;
static int             num_packets = 0;

static int
ones_complement_checksum(const void *p, int b, u_int32_t sum)
{
  const u_short *sp;

  sp = (u_short *) p; /* better be aligned! */
  b /= 2;             /* convert to count of short's */

  /* No need for endian conversions. */
  while ( --b >= 0 )
    sum += *sp++;

  while ( sum > 0xffff )
    sum = (sum & 0xffff) + (sum >> 16);

  return (sum);
}

static u_short
in_cksum(register u_short *addr, register int len, u_int preadd)
{
  register int nleft = len;
  register u_short *w = addr;
  register u_short answer;
  register int sum = 0;

  sum += preadd;

  /*
   *  Our algorithm is simple, using a 32 bit accumulator (sum),
   *  we add sequential 16 bit words to it, and at the end, fold
   *  back all the carry bits from the top 16 bits into the lower
   *  16 bits.
   */
  while (nleft > 1) {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nleft == 1)
    sum += *(u_char *)w;

  /*
   * add back carry outs from top 16 bits to low 16 bits
   */
  sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
  sum += (sum >> 16);                 /* add carry */
  answer = ~sum;                      /* truncate to 16 bits */

  return answer;
}


static u_int16_t 
ip_checksum(struct ip* iphdr)
{
  u_int16_t    old, result = 0;

  if (!iphdr)
    return (0);
  
  old = iphdr->ip_sum;
  iphdr->ip_sum = 0;
  result = in_cksum((u_short *)iphdr, iphdr->ip_hl<<2, 0);
  iphdr->ip_sum = old;

  return result;
}


static u_int16_t
icmp_checksum(struct ip *iphdr, struct icmp *icmphdr)
{
  u_int16_t      old, result = 0;

  if (!(iphdr && icmphdr))
    return (0);
  
  old = icmphdr->icmp_cksum;
  icmphdr->icmp_cksum = 0;
  result = in_cksum((u_short *)icmphdr, ntohs(iphdr->ip_len) - (iphdr->ip_hl<<2), 0);
  icmphdr->icmp_cksum = old;

  return result;
}


static u_int16_t 
tcp_checksum(struct ip* iphdr, struct tcphdr *tcphdr)
{
  int             tcp_len, len;
  u_int32_t       sum, addl_pseudo;
  u_int16_t       old, result = 0;

  if (!(iphdr && tcphdr))
    return (0);

  old = tcphdr->th_sum;
  tcphdr->th_sum = 0;
  
  /* len is payload length, tcp_len is length of entire tcp packet: */
  tcp_len = (ntohs(iphdr->ip_len) - (iphdr->ip_hl * 4));
  len = tcp_len - tcphdr->th_off*4;
  
  if (len % 2 == 1) /* Add in pad byte. */
    sum = htons(((const u_char*) tcphdr)[tcp_len - 1] << 8);
  else
    sum = 0;
  
  sum = ones_complement_checksum((u_short*) &iphdr->ip_src.s_addr, 4, sum);
  sum = ones_complement_checksum((u_short*) &iphdr->ip_dst.s_addr, 4, sum);
  
  addl_pseudo = (htons(IPPROTO_TCP) << 16) | htons((u_short) tcp_len);
  
  sum = ones_complement_checksum((u_short*) &addl_pseudo, 4, sum);
  sum = ones_complement_checksum((u_short*) tcphdr, tcp_len, sum);
  
  tcphdr->th_sum = old;
  result = ~(sum);
      
  return (result);
}


u_int16_t 
udp_checksum(struct ip *iphdr, struct udphdr *udphdr)
{
  struct ipovly   ipovly;
  u_short        *w;
  u_int           i, preadd;
  u_int16_t       old, result = 0;
  
  if (!(iphdr && udphdr))
    return (0);

  w  = (u_short*)&ipovly;
  
  preadd = 0;
      
  bzero(&ipovly, sizeof(struct ipovly));
  ipovly.ih_pr  = iphdr->ip_p;
  ipovly.ih_len = udphdr->uh_ulen;
  ipovly.ih_src = iphdr->ip_src;
  ipovly.ih_dst = iphdr->ip_dst;
  
  for (i = 0; i < sizeof(struct ipovly)/sizeof(u_short) ; i++)
    preadd += *w++;
  
  old = udphdr->uh_sum;
  udphdr->uh_sum = 0;
  result = in_cksum((u_short*)udphdr, ntohs(udphdr->uh_ulen), preadd);
  udphdr->uh_sum = old;

  return (result);
}


static void
adjust_checksums(const struct pcap_pkthdr* pkthdr, u_char * packet)
{
  struct ip      *iphdr   = NULL;
  struct tcphdr  *tcphdr  = NULL;
  struct udphdr  *udphdr  = NULL;
  struct icmp    *icmphdr = NULL;
  u_char         *offset  = NULL;

  switch (pcap_datalink(pcap))
    {
    case DLT_EN10MB:
      {
        struct ether_header *eh;
	struct snap_header  *sh;

	eh = (struct ether_header*)packet;
	
	/* Is this an Ethernet-II frame or a Ethernet 802.3 SNAP frame?*/
	/* Check ether_type > 1500 --> Ethernet II */
	if (htons(eh->ether_type) <= ETHERMTU)
	  {
	    sh = (struct snap_header*)(packet + ETHER_HDR_LEN);
	    
	    if (sh->ssap != 0xaa)
	        return;
	    
	    /* Check for IP payload: 0x00-00-00-08-00 SNAP field */
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) | 
		 (!sh->snap[4])) && (sh->snap[3]==8))
	      offset = packet + (ETHER_HDR_LEN + SNAP_LEN);
	  }
	else 
	  {
	    if (htons(eh->ether_type) == ETHERTYPE_IP)
	      offset = packet + (ETHER_HDR_LEN);
	    else
	      return;
	  }
      }
      break;

    case DLT_FDDI:
      {
	struct fddi_header *fh;
	struct snap_header *sh;

	fh = (struct fddi_header*)packet;
	sh = (struct snap_header*)(packet + FDDI_HDR_LEN);


	/*Is this an FDDI 802.2 or an FDDI SNAP frame?*/
	if ((sh->ssap==0xaa)&&(sh->dsap==0xaa))
	  {
	    /*It's a SNAP frame*/
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) |
		 (sh->snap[4]==0)) &&(sh->snap[3]==8))
	      /*It's a SNAP header containing an IP datagram*/
	      offset = packet + (FDDI_HDR_LEN + SNAP_LEN);
	    else
	      return;
	  }
	else
	  return;
      }
      break;

    default:
      offset = packet;
    }

  if (!offset)
    return;

  /* Offset is at IP hdr now: */
  iphdr = (struct ip*)offset;

  iphdr->ip_sum = ip_checksum(iphdr);

  acc_size += ntohs(iphdr->ip_len);

  if (((ntohs(iphdr->ip_off) & IP_OFFMASK) == 0)  &&
      !(ntohs(iphdr->ip_off) & IP_MF))
    {
      switch (iphdr->ip_p)
	{
	case IPPROTO_ICMP:
	  {
	    icmphdr = (struct icmp*)((u_char*)iphdr + 4*iphdr->ip_hl);
	    icmphdr->icmp_cksum = icmp_checksum(iphdr, icmphdr);
	  }
	  break;
	case IPPROTO_UDP:
	  {
	    udphdr = (struct udphdr*)((u_char*)iphdr + 4*iphdr->ip_hl);
	    udphdr->uh_sum = udp_checksum(iphdr, udphdr);
	  }
	  break;
	case IPPROTO_TCP:
	  {
	    tcphdr = (struct tcphdr*)((u_char*)iphdr + 4*iphdr->ip_hl);
	    tcphdr->th_sum = tcp_checksum(iphdr, tcphdr);
	  }
	  break;
	}
    }

  return;
  pkthdr = NULL;
}


static void
handler(u_char *user, const struct pcap_pkthdr* pkthdr, const u_char * packet)
{
  num_packets++;
  adjust_checksums(pkthdr, (u_char*)packet);
  pcap_dump((u_char*)dumper, pkthdr, packet);

  return;
  user = NULL;
}


int
main(int argc, char **argv)
{
  if (argc < 3)
    {
      printf("USAGE: %s <input trace> <output trace>\n", argv[0]);
      exit(1);
    }

  input  = argv[1];
  output = argv[2];

  pcap = pcap_open_offline(input, errbuf);
  dumper = pcap_dump_open(pcap, output);
  pcap_loop(pcap, 0, handler, NULL);
  pcap_close(pcap);
  pcap_dump_close(dumper);

  printf("%i packets, mean size %f\n", num_packets, (float)acc_size/(float)num_packets);

  return (0);
}
