#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include <norm.h>
#include <norm_list.h>
#include <norm_hash.h>
#include <norm_debug.h>



NormHash   *
norm_hash_new(NormHashFunc hash_func, NormCmpFunc cmp_func,
	      NormFunc free_func, int num_slots, int max_slot_load)
{
  NormHash *h;

  D_ENTER;
  h = NEW(NormHash);
  h->slots = (NormList**)malloc(sizeof(NormList*) * num_slots);
  h->slot_loads = (u_int*)malloc(sizeof(u_int) * num_slots);

  bzero(h->slots, num_slots * sizeof(NormList*));
  bzero(h->slot_loads, num_slots * sizeof(u_int));

  h->hash_func = hash_func;
  h->cmp_func = cmp_func;
  h->free_func = free_func;

  h->num_slots = num_slots;
  h->num_items = 0;

  if (max_slot_load > 0)
    h->max_slot_load = max_slot_load;
  else
    h->max_slot_load = UINT_MAX;

  D(NOISE_LOGLEVEL, ("Allocated Hashtable.\n"));	  
  D_RETURN_(h);
}


void        
norm_hash_free(NormHash* h)
{
  u_int i;

  D_ENTER;
  for (i = 0; i < h->num_slots; i++)
    {
      if (h->slots[i])
	{
	  norm_list_free(h->slots[i], h->free_func);
	  h->slots[i] = NULL;
	}
    }
  free(h);
  D(NOISE_LOGLEVEL, ("Freed Hashtable.\n"));	  
  D_RETURN;
}


void        
norm_hash_insert(NormHash* h, void *key, void *data)
{
  NormHashItem  *it;
  u_int index;

  D_ENTER;

  if (norm_hash_find(h, key))
    {
      D(NOISE_LOGLEVEL, ("Item exists -- not inserting.\n"));
      D_RETURN;
    }

  index = h->hash_func(h, key);

  it = NEW(NormHashItem);
  it->data = data;
  it->key  = key;

  if (h->slot_loads[index] + 1 == h->max_slot_load)
    {
      D(NOISE_LOGLEVEL, ("Hash slot overflow -- removing last item.\n"));
      h->slots[index] = norm_list_remove_last(h->slots[index], h->free_func);
      h->slot_loads[index]--;
      h->num_items--;      
    }
  
  h->slots[index] = norm_list_prepend(h->slots[index], it);
  h->slot_loads[index]++;
  h->num_items++;
  D(NOISE_LOGLEVEL, ("Inserted Item, now %i\n", h->num_items));	  
  D_RETURN;
}


NormHashItem*
norm_hash_find(NormHash *h, void *key)
{
  NormList  *l;
  u_int      index;
  int        i;

  D_ENTER;
  index = h->hash_func(h, key);

  for (i = 1, l = h->slots[index]; l; l = l->next, i++)
    {
      if (h->cmp_func(key, ((NormHashItem*)l->data)->key) == 0)
	{
	  D(NOISE_LOGLEVEL, ("Key with hash value %i found at %i. item\n", index, i));
	  h->slots[index] = norm_list_move_to_front(h->slots[index], l);
	  D_RETURN_(l->data);
	}
    }

  D(NOISE_LOGLEVEL, ("Key with hash value %i not found.\n", index));
  D_RETURN_(NULL);
}


void
norm_hash_remove(NormHash *h, void *key)
{
  NormList     *l = NULL;
  u_int         index;

  D_ENTER;
  index = h->hash_func(h, key);

  for (l = h->slots[index]; l; l = l->next)
    {
      if (h->cmp_func(key, ((NormHashItem*)l->data)->key) == 0)
	break;
    }

  if (l)
    {
      h->slots[index] = norm_list_remove(h->slots[index], l, h->free_func);
      h->slot_loads[index]--;
      h->num_items--;
      D(NOISE_LOGLEVEL, ("Removed Item, hashed %i, now %i items\n", index, h->num_items));	  
    }
  else
    {
      D(NOISE_LOGLEVEL, ("Item that hashed %i not removed, still %i items.\n", index, h->num_items));	  
    }
  D_RETURN;
}
