/* Contains code derived from traceroute, which has the following copyright:
 *
 * Copyright (c) 1988, 1989, 1991, 1994, 1995, 1996, 1997
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/time.h>
#include <errno.h>
 
#include <norm.h>
#include <norm_ip.h>
#include <norm_ip_fragment.h>
#include <norm_tcp.h>
#include <norm_debug.h>
#include <norm_hash.h>

#include <arpa/inet.h>


/* Static function prototypes: */
static NormFragment *norm_fragment_new(struct ip *iphdr, u_int len);
static void          norm_fragment_free(NormFragment *f);

static NormFPacket*  norm_fpacket_new(struct ip* iphdr);
static void          norm_fpacket_free(NormFPacket *fp);
static void          norm_fpacket_add_fragment(NormFPacket *fp, NormFragment *f,
					       const struct timeval *ts);

static void          norm_ip_fragment_free_oldest(void);
static u_int         norm_ip_fragment_hash_iphdr(NormHash* h, void *hdr);
static int           norm_ip_fragment_cmp_ip(void *ip1, void *ip2);
static int           norm_ip_fragment_attempt_reassembly(NormFPacket *fp, Interface *nif);

/* Hashtable containing all fragmented packets */
NormHash  *fp_hash = NULL;


static NormFragment*
norm_fragment_new(struct ip *iphdr, u_int len)
{
  u_int         hlen;
  NormFragment *f;

  D_ENTER;

  f = NEW(NormFragment);
  bzero(f, sizeof(NormFragment));
  f->data = (char*)malloc(len);
  
  hlen = iphdr->ip_hl*4;
  memcpy(f->data, (u_char*)iphdr + hlen, len - hlen);
  f->len = len - hlen;

  if ((htons(iphdr->ip_off) & IP_MF) == IP_MF)
    f->mf = 1;  
  else
    f->mf = 0;

  f->offset = (htons(iphdr->ip_off) & IP_OFFMASK) << 3;
  print_ip(INFO_LOGLEVEL, ("mf:%d offset: %d len %d\n", f->mf, f->offset, f->len));
  
  D_RETURN_(f);
}


static void
norm_fragment_free(NormFragment *f)
{
  D_ENTER;

  if (!f)
    D_RETURN;

  D(NOISE_LOGLEVEL, ("Freeing NormFragment.\n"));

  free(f->data);
  free(f);

  D_RETURN;
}


static NormFPacket*
norm_fpacket_new(struct ip* iphdr)
{
  NormFPacket *fp;

  D_ENTER;
  fp = NEW(NormFPacket);
  bzero(fp, sizeof(NormFPacket));
  
  fp->src = iphdr->ip_src;
  fp->dst = iphdr->ip_dst;
  fp->id  = iphdr->ip_id;

  fp->header = (char*)malloc(iphdr->ip_hl*4);
  memcpy(fp->header, iphdr, iphdr->ip_hl*4);
  fp->headerlen = iphdr->ip_hl*4;

  D_RETURN_(fp);
}
  

static void
norm_fpacket_free(NormFPacket *fp)
{
  D_ENTER;

  if (!fp)
    D_RETURN;

  norm_list_free(fp->fragments, (NormFunc)norm_fragment_free);

  if (fp->header)
    free(fp->header);
  free(fp);

  D_RETURN;
}


static void
norm_fpacket_add_fragment(NormFPacket *fp, NormFragment *f, const struct timeval *ts)
{
  NormList *l, *lp;

  D_ENTER;

  if (!fp || !f)
    return;

  fp->num_frags++;
  fp->fragsize+= f->len;
  fp->ts.tv_sec = ts->tv_sec;
  fp->ts.tv_usec = ts->tv_usec;

  /*add fragment to linked list of fragments in the correct sequence*/
  l = lp = fp->fragments;

  while(l && (((NormFragment*)l->data)->offset < f->offset))
    {
      lp = l; l = l->next;
    }

  if (!l && lp)
    {
      /*it's the last fragment so far*/
      D(NOISE_LOGLEVEL, ("Adding as last fragment\n"));
      norm_list_append(lp, f);
    }
  else if (l == fp->fragments)
    {
      /*it's the first fragment so far*/
      D(NOISE_LOGLEVEL, ("First fragment.\n"));
      fp->fragments = norm_list_prepend(fp->fragments, f);
    }
  else if (((NormFragment*)l->data)->offset == f->offset)
    {
      /*it's a duplicate fragment*/
      D(NOISE_LOGLEVEL, ("Duplicate fragment -- freeing()."));
      norm_fragment_free(f);
      D_RETURN;
    }
  else
    {
      /*it's somewhere in the middle*/
      D(NOISE_LOGLEVEL, ("Middle fragment\n"));
      norm_list_insert_after(lp, f);
    }
    

  if ((fp->num_frags > NORM_MAX_FRAGS_PER_PACKET) ||
      (fp->fragsize > 65535))
    {
      print_ip(FATAL_LOGLEVEL, ("max frags per packet exceeded!\n"));
      norm_hash_remove(fp_hash, fp->header);
      D_RETURN;
    } 
  
  D_RETURN;
}

static void
norm_ip_fragment_free_oldest(void)
{
  NormHashItem    *it;
  NormFPacket     *fp, *oldest_fp = NULL;
  struct timeval   oldtime;
  u_int            i, first = TRUE;

  D_ENTER;
  
  for (i = 0; i < fp_hash->num_slots; i++)
    {
      if (fp_hash->slots[i])
	{
	  it = (NormHashItem*)fp_hash->slots[i]->data;
	  fp = (NormFPacket*)it->data;
	  
	  if (first ||
	      ((fp->ts.tv_sec < oldtime.tv_sec)||
	       ((fp->ts.tv_sec == oldtime.tv_sec)&&
		(fp->ts.tv_usec < oldtime.tv_usec))))
	    {
	      oldest_fp = fp;
	      oldtime.tv_sec = fp->ts.tv_sec;
	      oldtime.tv_usec = fp->ts.tv_usec;
	      first = FALSE;
	    }
	}
    }
  
  if (oldest_fp)
    norm_hash_remove(fp_hash, oldest_fp->header);
  D_RETURN;
}


static u_int
norm_ip_fragment_hash_iphdr(NormHash* h, void *hdr)
{
  struct ip* iphdr = (struct ip*)hdr; 
  
  u_int16_t id  = iphdr->ip_id;
  u_int32_t src = iphdr->ip_src.s_addr;
  u_int32_t dst = iphdr->ip_dst.s_addr;
  u_int8_t  p   = iphdr->ip_p;

  D_ENTER;
    
  D_RETURN_(((id >> 1) ^ src ^ dst ^ p) & (h->num_slots - 1));
}


static int
norm_ip_fragment_cmp_ip(void *ip1, void *ip2)
{
  struct ip *iphdr1 = (struct ip*)ip1; 
  struct ip *iphdr2 = (struct ip*)ip2; 

  D_ENTER;

  if (iphdr1->ip_src.s_addr < iphdr2->ip_src.s_addr)
    D_RETURN_(-1);
  if (iphdr1->ip_src.s_addr > iphdr2->ip_src.s_addr)
    D_RETURN_(1);
  if (iphdr1->ip_dst.s_addr < iphdr2->ip_dst.s_addr)
    D_RETURN_(-1);
  if (iphdr1->ip_dst.s_addr > iphdr2->ip_dst.s_addr)
    D_RETURN_(1);
  if (iphdr1->ip_id < iphdr2->ip_id)
    D_RETURN_(-1);
  if (iphdr1->ip_id > iphdr2->ip_id)
    D_RETURN_(1);

  D_RETURN_(0);
}


static void
norm_ip_fragment_hash_item_free(void *data)
{
  NormHashItem *it = (NormHashItem*)data;

  D_ENTER;
  /* This case is special -- the it->data
     contains the it->key, so we don't need
     to free the key separately ...
  */

  /* free(it->key); */
  norm_fpacket_free(it->data);
  D_RETURN;
}


static int
norm_ip_fragment_attempt_reassembly(NormFPacket *fp, Interface *nif)
{
  NormFragment      *f, *f2;
  NormList          *l;
  struct ip         *iphdr;
  u_int              len=0, ctr=0;
  u_char            *ptr;
  NormPacket         p;

  D_ENTER;

  /*check if we have all the fragments*/

  l = fp->fragments;
  f = (NormFragment*)l->data;

  if (f)
    {
      if (f->offset != 0)
	{
	  print_ip(NOISE_LOGLEVEL, ("first frag has offset %d\n", f->offset));
	  D_RETURN_(-1);
	}
    }

  while (l->next)
    {
      f = (NormFragment*)l->data;
      f2 = (NormFragment*)l->next->data;

      if ((f->offset + f->len) > len)
	len = f->offset + f->len;

      print_ip(INFO_LOGLEVEL, ("frag %d off: %d len:%d tlen:%d\n", ctr++, f->offset, f->len, len));

      /*does this fragment match the next one*/
      /*note: it may overlap - we don't consider this an error*/

      if ((f->offset + f->len) < f2->offset)
	break;
      l = l->next;
    }

  f = (NormFragment*)l->data;

  if ((f->offset + f->len) > len)
    len = f->offset + f->len;

  print_ip(INFO_LOGLEVEL, ("frag %d off: %d len:%d tlen:%d\n", ctr++, f->offset, f->len, len));

  if (f->mf == 1)
    {
      print_ip(INFO_LOGLEVEL, ("last frag has MF set\n"));
      D_RETURN_(-1);
    }

  if (len + fp->headerlen > 65535)
    {
      print_ip(NORM_LOGLEVEL, ("max packet length exceeded: %d\n", len+fp->headerlen));
      norm_hash_remove(fp_hash, fp->header);
      D_RETURN_(-1);
    }

  print_ip(INFO_LOGLEVEL, ("got all fragments, len=%d\n", len));

  bzero(&p, sizeof(NormPacket));
  norm_packet_create(&p, len + fp->headerlen, nif);
  memcpy(p.network_data, fp->header, fp->headerlen);

  l = fp->fragments;
  while (l)
    {
      f = (NormFragment*)l->data;
      ptr = p.network_data + fp->headerlen + f->offset;
      memcpy(ptr, f->data, f->len);
      l = l->next;
  }

  iphdr = (struct ip*)p.network_data;
  iphdr->ip_len = htons(len + fp->headerlen);
  iphdr->ip_off = 0;
  iphdr->ip_sum = 0;
  iphdr->ip_sum = norm_ip_in_cksum((u_short *)iphdr, iphdr->ip_hl<<2, 0);

  D(NORM_LOGLEVEL, ("Injecting reassembled packet.\n"));
  norm_ip_handle_regular_packet(&p, nif);
  norm_empty_packet(&p);
  norm_hash_remove(fp_hash, fp->header);

  D_RETURN_(0);
}


void
norm_ip_fragment_init(void)
{
  D_ENTER;
  if (!fp_hash)
    {
      fp_hash = norm_hash_new(norm_ip_fragment_hash_iphdr,
			      norm_ip_fragment_cmp_ip,
			      norm_ip_fragment_hash_item_free,
			      NORM_MAX_FRAGGED_PACKETS,
			      NORM_MAX_FRAGGED_PACKETS);
    }
  D_RETURN;
}


int
norm_ip_fragment_handler(NormPacket *p, Interface *nif)
{
  static u_int       fp_size_max = 0;

  NormHashItem      *it;
  NormFragment      *f;
  NormFPacket       *fp;
  struct ip         *iphdr;
  int                offset;
  int                len;

  D_ENTER;

  iphdr  = (struct ip*)p->network_data;
  offset = (htons(iphdr->ip_off) & IP_OFFMASK) << 3;
  len    = ntohs(iphdr->ip_len);

  f = norm_fragment_new(iphdr, len);

  if ( (it = (norm_hash_find(fp_hash, iphdr))))
    {
      D(NOISE_LOGLEVEL, ("existing packet.\n"));
      
      fp = (NormFPacket*)it->data;
      
      norm_fpacket_add_fragment(fp, f, &(p->ph->ts));
      norm_ip_fragment_attempt_reassembly(fp, nif);
      D_RETURN_(0);
    }

  print_ip(NOISE_LOGLEVEL, ("new fragmented packet at offset %d.\n", offset));
  
  /* if we get here there are no fragments for this packet yet. */
  
  fp = norm_fpacket_new(iphdr);
  norm_fpacket_add_fragment(fp, f, &(p->ph->ts));
  
  norm_hash_insert(fp_hash, fp->header, fp);

  if (fp_hash->num_items > fp_size_max)
    {
      fp_size_max = fp_hash->num_items;
      printf("Larger hash size: %i items.\n", fp_size_max);
    }

  if (fp_hash->num_items > NORM_MAX_FRAGGED_PACKETS)
    {
      print_ip(FATAL_LOGLEVEL, ("max number of fragmented packets exceeded!\n"));
      norm_ip_fragment_free_oldest();
    }
 
  D_RETURN_(0);
}


