#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "norm.h"
#include "norm_misc.h"
#include "norm_options.h"
#include "norm_debug.h"

extern int  parse_init(char *filename);
extern void parse_cleanup(void);
extern int  norm_parse(void);

LogLevel loglevels[] =
{
  {   FATAL_LOGLEVEL, "FATAL"},
  {    NORM_LOGLEVEL, "NORM"},
  {    INFO_LOGLEVEL, "INFO"},
  { WARNING_LOGLEVEL, "WARNING"},
  {   NOISE_LOGLEVEL, "NOISE"},
  { NESTING_LOGLEVEL, "NESTING"},
  {                0, "GENERAL"}
};

NormOptions options;

static void  norm_options_show_help(char * name);
static void  norm_options_show_configuration(void);
static char *norm_options_get_configfile(void);
static void  norm_options_sighandler(int sig);

static void
norm_options_show_help(char * name)
{
  D_ENTER;
  printf("Usage: %s [options] ...\n"
	 "-h |-?|--help                    Display this information\n"
	 "-d |--debug-level <int>          Set debugging level (%i)\n"
	 "-s |--show-code-location         Show where in code debugging messages are (%i)\n"
	 "-ti|--in-tracefile <filename>    Input file for offline capture (%s)\n"
	 "-to|--out-tracefile <filename>   Output file for offline capture (%s)\n"
	 "-if|--interface <name>           Network interface name, can be given multiple times.\n"
	 "-c |--config-file <name>         Use <name> as config file ()\n" ,
	 name, options.debuglevel, options.show_code_location,
	 options.in_tracefile, options.out_tracefile);
  D_RETURN;
}


static void
norm_options_show_configuration(void)
{
  int  none;
  int  i;

  printf("=======================================================\n");
  printf("                     Norm setup:\n");
  printf("-------------------------------------------------------\n");
#ifdef DEBUG
  printf("Debug level: %i\n", options.debuglevel);
#endif
  if (options.passive)
    {
      printf("Running in PASSIVE mode\n");
      printf("=======================================================\n");
      return;
    }

  if (options.in_tracefile)
    {
      printf("Tracefile to read:  %s\n", options.in_tracefile);
      printf("Tracefile to write: %s\n", options.out_tracefile);
    }
  else
    {
      printf("\nNetwork interface(s): ");
      for (i = 0; i < options.num_ifs; i++)
	printf("%s ", options.ifs[i]);

      printf("\n");
    }

  for (i = 0; i < options.num_trusted; i++)
    {
      struct in_addr addr;

      addr.s_addr = htonl(options.trusted_net[i].s_addr);
      printf("Trusted network: %s\n", inet_ntoa(addr));
    }


  printf("\nProtocol output:\n");
  
  printf("Linklevel: ");
  none = TRUE;
  if (options.show_ether) { none = FALSE; printf("Ethernet "); }
  if (options.show_fddi)  { none = FALSE; printf("FDDI"); }
  if (none) printf("none");
  
  printf("\nNetwork:   ");
  none = TRUE;
  if (options.show_ip)  { none = FALSE; printf("IP "); }
  if (none) { printf("none"); }
  
  printf("\nTransport: ");
  none = TRUE;
  if (options.show_tcp)   { none = FALSE; printf("TCP "); }
  if (options.show_udp)   { none = FALSE; printf("UDP "); }
  if (options.show_icmp)  { none = FALSE; printf("ICMP "); }
  if (none) { printf("none"); }
  
  printf("\n\nNormalizations:\n");
  printf("IP:\n");
  if (options.ip_drop_fragments_with_df)      printf(" - Drop IP fragments with DF set.\n");
  if (options.ip_clear_df_and_offset)         printf(" - Clear DF flag and fragment offset.\n");
  if (options.ip_check_fragment_offset)       printf(" - Sanity-check IP fragments offset + packet length.\n");
  if (options.ip_check_options)               printf(" - Sanity-check IP options.\n");
  if (options.ip_header_length_check)         printf(" - Check IP header length.\n");
  if (options.ip_total_length_check)          printf(" - Check IP packet length.\n");
  if (options.ip_header_checksum_check)       printf(" - Validate IP header checksum.\n");
  if (options.ip_header_addr_check)           printf(" - Sanity-check IP source and dest addrs.\n");
  if (options.ip_header_clear_tos)            printf(" - Clear Type-of-Service flags.\n");
  if (options.ip_check_ttl)                   printf(" - Set TTL field to %i.\n", options.ip_ttl_value);
  if (options.ip_clear_rf)                    printf(" - Clear reserved fragment bit.\n");
  printf("UDP:\n");
  if (options.udp_verify_checksum)            printf(" - Verify UDP checksum.\n");
  if (options.udp_verify_length)              printf(" - Verify UDP packet length.\n");
  printf("TCP:\n");
  if (options.tcp_need_syn)                   printf(" - Need to see SYN on new connections.\n");
  if (options.tcp_verify_checksum)            printf(" - Verify TCP checksum.\n");
  if (options.tcp_check_flags)                printf(" - Check flag combinations.\n");
  if (options.tcp_check_lengths)              printf(" - Check packet size and data offset.\n");
  if (options.tcp_check_options)              printf(" - Check options.\n");
  if (options.tcp_clear_reserved)             printf(" - Clear reserved fields.\n");
  printf("ICMP:\n");
  if (options.icmp_drop_nonzero_code)         printf(" - Drop ICMP packets with non-zero code field.\n");
  if (options.icmp_drop_broadcast)            printf(" - Drop ICMP ping requests with broadcast src or dest.\n");
  if (options.icmp_drop_multicast)            printf(" - Drop ICMP ping requests with multicast dest.\n");
  if (options.icmp_drop_sourcequench)         printf(" - Drop ICMP source quench messages.\n");
  if (options.icmp_verify_checksum)           printf(" - Verify ICMP checksum.\n");
  printf("=======================================================\n");
}


static char *
norm_options_get_configfile(void)
{
  char *s;

  s = strdup(getenv("HOME"));
  s = realloc(s, strlen(s) + 7);
  strcat(s, "/.norm");

  if (access(s, R_OK) < 0)
    {
      printf("Error in configuration -- configfile not found in %s.\n", s);
      free(s);
      exit(1);
    }

  return s;
}


static void  
norm_options_sighandler(int sig)
{
  printf("Reloading configuration ...\n");

  /* clean up first */
  bzero(&FIRST_SWITCH, sizeof(SWITCH_TYPE) * NUM_SWITCHES);

  /* re-read config file: */
  if (parse_init(options.configfile))
    {
      norm_parse();
      parse_cleanup();
      norm_options_show_configuration();
    }
  
  return;
  sig = 0;
}


void
norm_options_read(int argc, char** argv)
{
  int i;
  int ok = TRUE;

  /* Set up default values ... */
  bzero(&options, sizeof(NormOptions));
  options.configfile = norm_options_get_configfile();

  /* Read config file: */
  if (parse_init(options.configfile))
    {
      norm_parse();
      parse_cleanup();
    }

  /* Override by command line options: */
  for (i = 0; i < argc; i++)
    {
      if ((!strcmp(argv[i], "--help")) || (!strcmp(argv[i], "-h"))
	  || (!strcmp(argv[i], "-?")))
	{
	  norm_options_show_help(argv[0]);
	  exit(0);
	}
      else if ((!strcmp(argv[i], "-v")) || (!strcmp(argv[i], "--verbose")))
	{
	  options.verbose = TRUE;
	}
      else if ((!strcmp(argv[i], "-d")) || (!strcmp(argv[i], "--debug-level")))
	{
	  int dl;

	  if ((i+1 < argc) && ((dl = atoi(argv[i+1])) >= 0))
	    options.debuglevel = dl;
	  i++;
	}
      else if ((!strcmp(argv[i], "-s")) || (!strcmp(argv[i], "--show-code-location")))
	{
	  options.show_code_location = TRUE;
	}
      else if ((!strcmp(argv[i], "-ti")) || (!strcmp(argv[i], "--in-tracefile")))
	{
	  if (i+1 < argc)
	    options.in_tracefile = argv[i+1];
	  i++;
	}
      else if ((!strcmp(argv[i], "-to")) || (!strcmp(argv[i], "--out-tracefile")))
	{
	  if (i+1 < argc)
	    options.out_tracefile = argv[i+1];
	  i++;
	}
      else if ((!strcmp(argv[i], "-if")) || (!strcmp(argv[i], "--interface")))
	{
	  if (i+1 < argc)
	    norm_misc_add_network_interface(argv[i+1]);
	  i++;
	}
      else if ((!strcmp(argv[i], "-c")) || (!strcmp(argv[i], "--config-file")))
	{
	  if (i+1 < argc)
	    {
	      if (options.configfile)
		FREE(options.configfile);
	    }
	    options.configfile = argv[i+1];
	  i++;
	}
    }

  /* sanity checks -- do we have everything we need for basic operation? */
  if (options.num_ifs == 0)
    {
      printf("No interfaces specified.\n");
      ok = FALSE;
    }
  if (options.in_tracefile && !options.out_tracefile)
    {
      printf("No output tracefile specified.\n");
      ok = FALSE;
    }
  if (options.out_tracefile && !options.in_tracefile)
    {
      printf("No input tracefile specified.\n");
      ok = FALSE;
    }

  if (!ok)
    {
      printf("Error in configuration -- aborting.\n");
      exit(1);
    }

  if (options.verbose)
    norm_options_show_configuration();
}


void   
norm_options_install_sighandler(void)
{
  signal(SIGHUP, norm_options_sighandler);
}

char *
norm_options_get_dbg_id(int loglevel)
{
  int i;

  i = 0;
  while (loglevels[i].level != 0)
    {
      if (loglevels[i].level == loglevel)
	return loglevels[i].id;
      i++;
    }
  
  return loglevels[i].id;
}
