/* checksum code derived from traceroute, which has the following copyright:
 *
 * Copyright (c) 1988, 1989, 1991, 1994, 1995, 1996, 1997
 *The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <norm.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <net/ethernet.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>

#include <norm_ip.h>
#include <norm_tcp.h>
#include <norm_misc.h>
#include <norm_debug.h>
#include <norm_options.h>
#include <norm_send.h>
#include <norm_tcb.h>
#include <norm_stats.h>


int
norm_tcp_checksum(const struct ip *ip, const struct tcphdr *tp, int len)
{
  int tcp_len;
  u_int32_t sum, addl_pseudo;

  D_ENTER;

  /* len is payload length, tcp_len is length of entire tcp packet: */
  tcp_len = tp->th_off * 4 + len;

  if ( len % 2 == 1 ) /* Add in pad byte. */
    sum = htons(((const u_char*) tp)[tcp_len - 1] << 8);
  else
    sum = 0;

  sum = norm_misc_ones_complement_checksum((void*) &ip->ip_src.s_addr, 4, sum);
  sum = norm_misc_ones_complement_checksum((void*) &ip->ip_dst.s_addr, 4, sum);

  addl_pseudo = (htons(IPPROTO_TCP) << 16) | htons((u_short) tcp_len);

  sum = norm_misc_ones_complement_checksum((void*) &addl_pseudo, 4, sum);
  sum = norm_misc_ones_complement_checksum((void*) tp, tcp_len, sum);

  D_RETURN_(sum);
}


int
norm_tcp_make_reset_reliable(struct ip *ih, struct tcphdr *th, int len, Interface *nif)
{
  /* we need to copy the packet, clear the flags, and queue it for sending
     after the actual RST was sent*/

  /* note: if the packet had any data, it's already been removed by now */
  NormPacket p;

  struct ip *new_ih;
  struct tcphdr *new_th;

  D_ENTER;

  bzero(&p, sizeof(NormPacket));
  norm_packet_create(&p, len, nif);

  new_ih = (struct ip *)p.network_data;
  memcpy(new_ih, ih, len);

  new_th = (struct tcphdr *)(((char*)new_ih) + ((u_int)th - (u_int)ih));

  new_th->th_flags = TH_ACK;
  new_th->th_sum=0;
  new_th->th_sum = ~norm_tcp_checksum(new_ih, new_th, 0);
  new_ih->ip_sum=0;
  new_ih->ip_sum = norm_ip_in_cksum((u_short*) new_ih, new_ih->ip_hl*4, 0);

  /* Add the packet to the send queue.  The queue will get flushed 
     immediately after ip_send_packet has sent the current packet */

  norm_packet_queue_for_sending(&p, nif);

  D_RETURN_(0);
}


int
norm_tcp_handler(NormPacket *p, Interface *nif)
{
  struct ip       *iphdr;
  struct tcphdr   *tcphdr;
  /* struct servent  *srvnt = NULL; */
  u_short          sum, mysum, sport, dport, window, urgent;
  u_int32_t        seqno, ackno;
  int              ip_len, tcp_data_len;

  /* for storing packets */
  NormTCB *tcb;

  D_ENTER;

  iphdr     = (struct ip*)p->network_data;
  tcphdr    = (struct tcphdr*)p->transport_data;

  sport  = ntohs(tcphdr->th_sport);
  dport  = ntohs(tcphdr->th_dport);
  window = ntohs(tcphdr->th_win);
  urgent = ntohs(tcphdr->th_urp);
  seqno  = ntohl(tcphdr->th_seq);
  ackno  = ntohl(tcphdr->th_ack);

  ip_len = norm_packet_get_length_ip(p);
  tcp_data_len = ip_len - (iphdr->ip_hl*4) - (tcphdr->th_off * 4);

  /* Sanity-check various lengths */
  if (options.tcp_check_lengths)
    {
      if (ip_len < (iphdr->ip_hl*4 + (int)sizeof(struct tcphdr)))
	{
	  norm_stats_increment_tcp_norms();
	  print_tcp(NORM_LOGLEVEL, ("packet is too small to be a TCP packet -- dropping.\n"));
	  D_RETURN_(-1);
	}
      
      if (ip_len < (iphdr->ip_hl*4 + tcphdr->th_off * 4))
	{
	  norm_stats_increment_tcp_norms();
	  print_tcp(NORM_LOGLEVEL, ("packet is too small for stated TCP data offset, dropping.\n"));
	  D_RETURN_(-1);
	}
      
      if (tcphdr->th_off < 5)
	{
	  norm_stats_increment_tcp_norms();
	  print_tcp(NORM_LOGLEVEL, ("TCP data offset too small, dropping.\n"));
	  D_RETURN_(-1);
	}
    }


  /*Check TCP checksum*/
  if (options.tcp_verify_checksum)
    {
      sum = tcphdr->th_sum;
      tcphdr->th_sum=0;
      
      mysum = ~norm_tcp_checksum(iphdr, tcphdr, tcp_data_len);
      if(sum != mysum)
	{
	  norm_stats_increment_tcp_norms();
	  print_tcp(NORM_LOGLEVEL, ("bad TCP checksum: %d should be %d, dropping.\n", htons(sum), htons(mysum)));
	  D_RETURN_(-1);
	}
      tcphdr->th_sum = sum;
    }
  
  /*The reserved fields should be zero*/
  if (options.tcp_clear_reserved && tcphdr->th_x2 != 0)
    {
      norm_stats_increment_tcp_norms();
      print_tcp(NORM_LOGLEVEL, ("packet with non-zero reserved field received, cleared.\n"));
      tcphdr->th_x2 = 0;
      tcphdr->th_flags &= 0x3f;
      p->transport_data_recalc_checksum = 1;
    }
  
  tcb = norm_tcb_lookup(iphdr, tcphdr, nif);
  
  if (options.show_tcp)
    {
      if (!tcb)
	{
	  print_tcp(INFO_LOGLEVEL, ("new:(*) payload:%d %i", tcp_data_len, sport));
	  
	  /*
	  srvnt = getservbyport(sport, "tcp");
	  if (srvnt)
	    {
	      print_tcp_(INFO_LOGLEVEL, ("(%s)", srvnt->s_name));
	    }
	  */

	  print_tcp_(INFO_LOGLEVEL, ("->%i", dport));

	  /*
	  srvnt = getservbyport(dport, "tcp");
	  if (srvnt)
	    {
	      print_tcp_(INFO_LOGLEVEL, ("(%s)", srvnt->s_name));
	    }
	  */

	  print_tcp_(INFO_LOGLEVEL, (" seq: %u ack:%u win:%d urp%d\n",
				     seqno, ackno, window, urgent));
	}
      else
	{
	  int offset_ack, offset_seq;

	  if (norm_misc_is_trusted_address(iphdr->ip_src))
	    {
	      offset_seq = seqno - tcb->first_int_seq_no;
	      offset_ack = ackno - tcb->first_ext_seq_no;
	    }
	  else
	    {
	      offset_seq = seqno - tcb->first_ext_seq_no;
	      offset_ack = ackno - tcb->first_int_seq_no;
	    }

	  print_tcp(INFO_LOGLEVEL, ("new:( ) payload:%d %i", tcp_data_len, sport));

	  /*
	  srvnt = getservbyport(tcphdr->th_sport, "tcp");
	  if (srvnt)
	    {
	      print_tcp_(INFO_LOGLEVEL, ("(%s)", srvnt->s_name));
	    }
	  */
	  print_tcp_(INFO_LOGLEVEL, ("->%i", dport));
	  /*
	  srvnt = getservbyport(tcphdr->th_dport, "tcp");
	  if (srvnt)
	    {
	      print_tcp_(INFO_LOGLEVEL, ("(%s)", srvnt->s_name));
	    }
	  */
	  print_tcp_(INFO_LOGLEVEL, (" seq:%u(+%u) ack:%u(+%u) win:%d urp:%d\n",
				     seqno, offset_seq, ackno, offset_ack, window, urgent));
	}
    }
    
  print_tcp(INFO_LOGLEVEL, ("URG(%c)  ACK(%c)  PSH(%c)  RST(%c)  SYN(%c)  FIN(%c)\n",
	    (tcphdr->th_flags&TH_URG ? '*' : ' '),  (tcphdr->th_flags&TH_ACK ? '*' : ' '),
	    (tcphdr->th_flags&TH_PUSH ? '*' : ' '), (tcphdr->th_flags&TH_RST ? '*' : ' '),
	    (tcphdr->th_flags&TH_SYN ? '*' : ' '),  (tcphdr->th_flags&TH_FIN ? '*' : ' ')));

  /* Check flags of desired: */
  if (options.tcp_check_flags)
    {
      if ((tcphdr->th_flags & TH_SYN) != 0)
	{
	  /* can't have RST set on a SYN segment */
	  if ((tcphdr->th_flags&TH_RST) !=0 )
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Bogus flags combination: SYN and RST\n"));
	      D_RETURN_(-1);
	    }

	  /* having SYN and FIN set isn't completely illegal, but we're
	     not going to let it pass - clear the FIN.  If this was a legit
	     use, the FIN will be resent anyway */
	  if((tcphdr->th_flags & TH_FIN) != 0)
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("SYN|FIN combination - clearing FIN\n"));
	      tcphdr->th_flags -= TH_FIN;
	      p->transport_data_recalc_checksum = 1;
	    }

	  if (tcp_data_len > 0)
	    {
	      /*this is a SYN packet containing data*/
	      /*we remove the data*/
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Removing data from SYN segment\n"));
	      ip_len -= tcp_data_len;
	      tcp_data_len = 0;
	      iphdr->ip_len = htons((iphdr->ip_hl + tcphdr->th_off) * 4);
	      p->transport_data_recalc_checksum = 1;
	    }
	}

      if ((tcphdr->th_flags & TH_FIN) != 0)
	{
	  /*can't have both FIN and RST*/
	  if((tcphdr->th_flags & TH_RST) != 0)
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Bogus flags combination: FIN and RST -- dropping.\n"));
	      D_RETURN_(-1); 
	    }
	}

      if ((tcphdr->th_flags & TH_RST) != 0)
	{
	  if (tcp_data_len > 0)
	    {
	      /*this is a RST packet containing data*/
	      /*we remove the data because the receiver should discard it anyway*/
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Removing data from RST segment\n"));
	      ip_len -= tcp_data_len;
	      tcp_data_len = 0;
	      iphdr->ip_len = htons((iphdr->ip_hl + tcphdr->th_off) * 4);
	      p->transport_data_recalc_checksum = 1;
	    }
	  
	  if (!norm_misc_is_trusted_address(iphdr->ip_src))
	    {
	      /* we can't trust the reset - need to make the RST reliable by
		 causing the trusted host to send a RST */
	      norm_stats_increment_tcp_norms();
	      norm_tcp_make_reset_reliable(iphdr, tcphdr, ip_len, nif);
	    }
	}

      if ((tcphdr->th_flags&TH_PUSH) != 0)
	{
	  if ((tcp_data_len==0) && ((tcphdr->th_flags & TH_FIN)==0) && ((tcphdr->th_flags & TH_RST)==0))
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Can't have PUSH with no data - clearing PUSH\n"));
	      tcphdr->th_flags -= TH_PUSH;
	      p->transport_data_recalc_checksum = 1;
	    }
	}
      
      if ((tcphdr->th_flags & TH_URG) != 0)
	{
	  /*check urgent pointer is sane*/
	  if (urgent > tcp_data_len)
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL,
			("Urgent pointer beyond data in packet - clearing flag and pointer.\n"));
	      tcphdr->th_flags -= TH_URG;
	      tcphdr->th_urp = 0;
	      p->transport_data_recalc_checksum = 1;
	    }
	}
      else
	{
	  /*check urgent pointer is zero - zero it if not*/
	  if (urgent != 0)
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Non-zero urgent pointer with URG not set!\n"));
	      tcphdr->th_urp = 0;
	      p->transport_data_recalc_checksum = 1;
	    }
	}

      /*ACK must be set on all segments except the initial SYN segment*/
      if ((tcphdr->th_flags & TH_ACK) == 0)
	{
	  if (((tcphdr->th_flags & TH_FIN)  != 0)  ||
#if 0
	      ((tcphdr->th_flags & TH_RST)  != 0)  ||
#endif
	      ((tcphdr->th_flags & TH_PUSH) != 0)  ||
	      ((tcphdr->th_flags & TH_URG)  != 0)  ||
	      ((tcphdr->th_flags & TH_SYN)  == 0))
	    {
	      norm_stats_increment_tcp_norms();
	      print_tcp(NORM_LOGLEVEL, ("Bogus flags combination: no ACK -- dropping.\n"));
	      D_RETURN_(-1);
	    }
	}
    }

  /* process options if desired */
  if (options.tcp_check_options)
    {
      int            optslen=0, optsdone=0, olen;
      int            removed=0; /*the number of bytes in options that we removed*/
      u_char        *optp;
      
      /*option info*/
      int mss=0;
      int wscale=0;

      optp = ((char*)tcphdr) + sizeof(struct tcphdr);
      optslen = (tcphdr->th_off)*4 - sizeof(struct tcphdr);
      optsdone = 0;

      while (optsdone < optslen)
	{
	  switch(*optp)
	    {
	    case TCPOPT_EOL:
	      removed = optslen + 1 - optsdone;
	      optsdone=optslen;
	      break;
	    case TCPOPT_NOP:
	      optp++;
	      optsdone++;
	      break;
	    case TCPOPT_MAXSEG:
	      /*MSS is only allowed on SYN segments*/
	      if ((tcphdr->th_flags&TH_SYN)==0)
		{
		  print_tcp(NORM_LOGLEVEL, ("Removing MSS option from non-SYN segment\n"));
		  olen = norm_tcp_remove_option(tcphdr,optp);
		  removed+=olen;
		  optsdone+=olen;
		  p->transport_data_recalc_checksum = 1;
		  break;
		}
	      optp++;
	      olen=*optp;
	      if (olen!=TCPOLEN_MAXSEG)
		{
		  print_tcp(NORM_LOGLEVEL, ("Incorrect TCPOPT_MAXSEG length: %d\n", olen));
		  D_RETURN_(-1);
		}
	      optp++;
	      mss=(int)ntohs(*(u_short*)optp);
	      print_tcp(INFO_LOGLEVEL, ("MSS: %d\n", mss));
	      optp+=olen-2;
	      optsdone+=olen;
	      break;
	    case TCPOPT_WINDOW:
	      /*WSCALE is only allowed on SYN segments*/
	      if ((tcphdr->th_flags&TH_SYN)==0) {
		print_tcp(NORM_LOGLEVEL, ("removing WSCALE option from non-SYN segment\n"));
		olen = norm_tcp_remove_option(tcphdr,optp);
		removed+=olen;
		optsdone+=olen;
		p->transport_data_recalc_checksum = 1;
		break;
	      }
	      optp++;
	      olen=*optp;
	      if (olen!=TCPOLEN_WINDOW)
		{
		  print_tcp(NORM_LOGLEVEL, ("Incorrect TCPOPT_WINDOW length: %d\n", olen));
		  D_RETURN_(-1);
		}
	      optp++;
	      wscale=(int)(*(u_char*)optp);
	      print_tcp(INFO_LOGLEVEL, ("WScale: %d\n", wscale));
	      optp+=olen-2;
	      optsdone+=olen;
	      break;
	    case TCPOPT_SACK_PERMITTED:
	      /*SACK_PERMITTED is only allowed on SYN segments*/
	      if ((tcphdr->th_flags&TH_SYN)==0) {
		print_tcp(NORM_LOGLEVEL, ("removing SACK_PERMITTED option from non-SYN segment\n"));
		olen = norm_tcp_remove_option(tcphdr,optp);
		removed+=olen;
		optsdone+=olen;
		p->transport_data_recalc_checksum = 1;
		break;
	      }
	      optp++;
	      olen=*optp;
	      if (olen!=TCPOPT_SACK_PERMITTED) {
		print_tcp(NORM_LOGLEVEL, ("Incorrect TCPOPT_SACK_PERMITTED length: %d\n", olen));
		D_RETURN_(-1);
	      }
	      print_tcp(INFO_LOGLEVEL, ("sack permitted option present\n"));
	      optp+=olen-1;
	      optsdone+=olen;
	      break;
	    case TCPOPT_SACK:
	      print_tcp(INFO_LOGLEVEL, ("sack option present\n"));
	      optp++;
	      olen=*optp;
	      optp+=olen-1;
	      optsdone+=olen;
	      break;
	    case TCPOPT_TIMESTAMP:
	      print_tcp(INFO_LOGLEVEL, ("timestamp option present\n"));
	      optp++;
	      olen=*optp;
	      optp+=olen-1;
	      optsdone+=olen;
	      break;
	    case TCPOPT_CC:
	      print_tcp(INFO_LOGLEVEL, ("CC option present\n"));
#ifdef ALLOW_TTCP
	      optp++;
	      olen=*optp;
	      optp+=olen-1;
	      optsdone+=olen;
#else
	      olen = norm_tcp_remove_option(tcphdr,optp);
	      removed+=olen;
	      optsdone+=olen;
	      p->transport_data_recalc_checksum = 1;
#endif
	      break;
	    case TCPOPT_CCNEW:
	      print_tcp(INFO_LOGLEVEL, ("CCNEW option present\n"));
#ifdef ALLOW_TTCP
	      optp++;
	      olen=*optp;
	      optp+=olen-1;
	      optsdone+=olen;
#else
	      olen = norm_tcp_remove_option(tcphdr,optp);
	      removed+=olen;
	      optsdone+=olen;
	      p->transport_data_recalc_checksum = 1;
#endif
	      break;
	    case TCPOPT_CCECHO:
	      print_tcp(INFO_LOGLEVEL, ("CCECHO option present\n"));
#ifdef ALLOW_TTCP
	      optp++;
	      olen=*optp;
	      optp+=olen-1;
	      optsdone+=olen;
#else
	      olen = norm_tcp_remove_option(tcphdr,optp);
	      removed+=olen;
	      optsdone+=olen;
	      p->transport_data_recalc_checksum = 1;
#endif
	      break;
	    case 19/*MD5 SIGNATURE*/:
	      /*we have to pass this or routing protocols may break!*/
	      /*XXXX - we need to not apply any other normalizations if this
		option is present*/
	      print_tcp(INFO_LOGLEVEL, ("MD5 signature option present\n"));
	      optp++;
	      olen=*optp;
	      optp+=olen-1;
	      optsdone+=olen;
	      break;
	    default:
	      print_tcp(NORM_LOGLEVEL, ("unknown TCP option present: %d\n", *optp));
	      olen = norm_tcp_remove_option(tcphdr,optp);
	      removed+=olen;
	      optsdone+=olen;
	      p->transport_data_recalc_checksum = 1;
	      break;
	    }
	  
	  if (optsdone>optslen)
	    {
	      print_tcp(NORM_LOGLEVEL, ("encountered bogus variable length option\n"));
	      D_RETURN_(-1);
	    }
	}

      if (removed >= 4)
	{
	  norm_stats_increment_tcp_norms();
	  removed = norm_tcp_fix_options(tcphdr, tcp_data_len, removed);
	  ip_len -= removed;
	  iphdr->ip_len = htons(ip_len);
	  p->transport_data_recalc_checksum = 1;
	}
      
      /*recalculate the IP and TCP checksums if we changed the packet*/
      if (p->transport_data_recalc_checksum)
	{	  
	  /*TCP checksum*/
	  tcphdr->th_sum = 0;
	  tcphdr->th_sum = ~norm_tcp_checksum(iphdr, tcphdr, tcp_data_len);
	  
	  /*IP header checksum (because total length may have changed)*/
	  iphdr->ip_sum = 0;
	  iphdr->ip_sum = norm_ip_in_cksum((u_short*)iphdr, iphdr->ip_hl*4, 0);

	  p->network_data_recalc_checksum = FALSE;
	  p->transport_data_recalc_checksum = FALSE;
	}
    }

  /*
   * If we got this far, this looks like a packet we're probably 
   * going to forward as it checked out (possibly with modifications) 
   * when viewed standalone.  Now we need to check if it makes sense 
   * when viewed as part of a connection.
   */

  /*Only create/use a TCB if this isn't a RST*/
  if ((tcphdr->th_flags & TH_RST) == 0)
    {
      if (!tcb)
	{
	  tcb = norm_tcb_new(iphdr, tcphdr, nif);
	  norm_tcb_store(tcb);
	}
      else
	{
	  if (norm_tcb_update(tcb, p, nif) < 0)
	    return (-1);
	}
    }

  if (tcb)
    {
      /* Update status of connection */

      if (norm_misc_is_trusted_address(iphdr->ip_src))
	{
	  if (tcphdr->th_flags & TH_SYN)
	    tcb->int_syn_seen = TRUE;
	  if (tcphdr->th_flags & TH_FIN)
	    tcb->int_fin_seen = TRUE;
	  if (tcphdr->th_flags & TH_RST)
	    tcb->int_rst_seen = TRUE;
	}
      else
	{
	  if (tcphdr->th_flags & TH_SYN)
	    tcb->ext_syn_seen = TRUE;
	  if (tcphdr->th_flags & TH_FIN)
	    tcb->ext_fin_seen = TRUE;
	  if (tcphdr->th_flags & TH_RST)
	    tcb->ext_rst_seen = TRUE;
	}

 
      /* Avoid stealth scan as described in 3.1.5:     */
      /* If there's a reset from the internal side and */
      /* no SYN was seen from the outside ...          */
      
      /* FIXME
	 if (nif->dir == INTERNAL_IF && tcphdr->th_flags & TH_RST && !tcb->ext_syn_seen)
	 {
	 }
      */

      /* Release this tcb if the connection is terminated. */
      if (tcb->int_rst_seen ||
	  (tcb->int_fin_seen && tcb->ext_fin_seen && (tcphdr->th_flags & TH_ACK) &&
	   !(tcphdr->th_flags & (TH_SYN|TH_FIN|TH_RST|TH_URG))))
	{
	  D(NOISE_LOGLEVEL, ("End of connection -- tearing down state.\n"));
	  norm_tcb_flush(tcb);
	}
    }

  D_RETURN_(0);
}


/* remove_option is called to remove the option starting at optp */
int
norm_tcp_remove_option(struct tcphdr *tcphdr, u_char *optp)
{
  int olen, thlen, remaining;

  D_ENTER;
  thlen = tcphdr->th_off*4;
  olen = *(optp+1);
  

  /*the remaining bytes after this option in the option chain*/
  remaining = thlen - (olen + (optp-(u_char*)tcphdr));

  /*copy the options chain up to remove this option*/
  if(remaining > 0)
    memmove(optp, optp+olen, remaining);

  if(remaining < 0) {
    print_tcp(NORM_LOGLEVEL, ("bogus TCP option length (longer than header!)\n"));
    D_RETURN_(olen);
  }

  /*Zero out the last olen bytes of the options chain.                     */
  /*This leaves the packet legal again, but perhaps lots of no-op options. */
  /*We remove all the no-ops in one pass at the end one we've removed all  */
  /*the other options we might want to remove*/
  bzero((u_char*)tcphdr+thlen-olen, olen);

  print_tcp(NORM_LOGLEVEL, ("%d bytes of TCP option removed\n", olen));
  D_RETURN_(olen);
}


/*fix_options is called when we removed more that four bytes of 
  TCP options. It removes the unnecessary no-op options that we added by
  tcp_remove_option */
int
norm_tcp_fix_options(struct tcphdr *tcphdr, int tlen, int remove)
{
  u_char    *eoopts;
  int        words;

  D_ENTER;

  print_tcp(INFO_LOGLEVEL, ("fix_up_tcp_options: %d bytes to be removed\n", remove));

  words = (remove >> 2);

  if (words == 0)
    D_RETURN_(0);

  remove = (words<<2);
  print_tcp(INFO_LOGLEVEL, ("fix_up_tcp_options: actually removing %d bytes\n", remove));

  eoopts  = (u_char*)tcphdr + tcphdr->th_off*4;

  if (tlen > 0)
    memmove(eoopts - remove, eoopts, tlen);

  tcphdr->th_off -= words;

  D_RETURN_(remove);
}




