#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <norm.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <net/ethernet.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>

#ifdef FREEBSD
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#endif
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>
#include <pcap.h>

#ifdef LINUX
struct ipovly
{
  u_char           ih_x1[9];
  u_char           ih_pr;
  u_int16_t        ih_len;
  struct in_addr   ih_src;
  struct in_addr   ih_dst;
};
#endif

#define FDDI_HDR_LEN    13
#define SNAP_LEN        8

static char             errbuf[PCAP_ERRBUF_SIZE];
static pcap_t         *pcap;
static char           *input;
static char           *output;
static pcap_dumper_t  *dumper;   

static int             too_small = 0;
static int             changed   = 0;
static int             fragmented = 0;
static int             unchanged = 0;
static int             not_ip    = 0;
static int             total = 0;
static int             t_udp       = 0;
static int             t_tcp       = 0;
static int             t_icmp      = 0;
static int             t_igmp      = 0;
static int             t_esp      = 0;
static int             t_pim      = 0;
static int             t_else      = 0;

static u_short
in_cksum(register u_short *addr, register int len, u_int preadd)
{
  register int nleft = len;
  register u_short *w = addr;
  register u_short answer;
  register int sum = 0;

  sum += preadd;

  /*
   *  Our algorithm is simple, using a 32 bit accumulator (sum),
   *  we add sequential 16 bit words to it, and at the end, fold
   *  back all the carry bits from the top 16 bits into the lower
   *  16 bits.
   */
  while (nleft > 1) {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nleft == 1)
    sum += *(u_char *)w;

  /*
   * add back carry outs from top 16 bits to low 16 bits
   */
  sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
  sum += (sum >> 16);                 /* add carry */
  answer = ~sum;                      /* truncate to 16 bits */

  return answer;
}


static u_int16_t 
ip_checksum(struct ip* iphdr)
{
  u_int16_t    old, result = 0;

  if (!iphdr)
    return (0);
  
  old = iphdr->ip_sum;
  iphdr->ip_sum = 0;
  result = in_cksum((u_short *)iphdr, iphdr->ip_hl<<2, 0);
  iphdr->ip_sum = old;

  return result;
}



u_int16_t 
udp_checksum(struct ip *iphdr, struct udphdr *udphdr)
{
  struct ipovly   ipovly;
  u_short        *w;
  u_int           i, preadd;
  u_int16_t       old, result = 0;
  
  if (!(iphdr && udphdr))
    return (0);

  w  = (u_short*)&ipovly;
  
  preadd = 0;
      
  bzero(&ipovly, sizeof(struct ipovly));
  ipovly.ih_pr  = iphdr->ip_p;
  ipovly.ih_len = udphdr->uh_ulen;
  ipovly.ih_src = iphdr->ip_src;
  ipovly.ih_dst = iphdr->ip_dst;
  
  for (i = 0; i < sizeof(struct ipovly)/sizeof(u_short) ; i++)
    preadd += *w++;
  
  old = udphdr->uh_sum;
  udphdr->uh_sum = 0;
  result = in_cksum((u_short*)udphdr, ntohs(udphdr->uh_ulen), preadd);
  udphdr->uh_sum = old;

  return (result);
}


static void
adjust_header(const struct pcap_pkthdr* pkthdr, u_char * packet)
{
  struct ip      *iphdr   = NULL;
  struct tcphdr  *tcphdr  = NULL;
  struct udphdr  *udphdr  = NULL;
  u_char         *offset  = NULL;

  total++;

  switch (pcap_datalink(pcap))
    {
    case DLT_EN10MB:
      {
        struct ether_header *eh;
	struct snap_header  *sh;

	eh = (struct ether_header*)packet;
	
	/* Is this an Ethernet-II frame or a Ethernet 802.3 SNAP frame?*/
	/* Check ether_type > 1500 --> Ethernet II */
	if (htons(eh->ether_type) <= ETHERMTU)
	  {
	    sh = (struct snap_header*)(packet + ETHER_HDR_LEN);
	    
	    if (sh->ssap != 0xaa)
	      {
		not_ip++;
		return;
	      }
	    
	    /* Check for IP payload: 0x00-00-00-08-00 SNAP field */
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) | 
		 (!sh->snap[4])) && (sh->snap[3]==8))
	      offset = packet + (ETHER_HDR_LEN + SNAP_LEN);
	  }
	else 
	  {
	    if (htons(eh->ether_type) == ETHERTYPE_IP)
	      offset = packet + (ETHER_HDR_LEN);
	  }
      }
      break;

    case DLT_FDDI:
      {
	struct fddi_header *fh;
	struct snap_header *sh;

	fh = (struct fddi_header*)packet;
	sh = (struct snap_header*)(packet + FDDI_HDR_LEN);


	/*Is this an FDDI 802.2 or an FDDI SNAP frame?*/
	if ((sh->ssap==0xaa)&&(sh->dsap==0xaa))
	  {
	    /*It's a SNAP frame*/
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) |
		 (sh->snap[4]==0)) &&(sh->snap[3]==8))
	      /*It's a SNAP header containing an IP datagram*/
	      offset = packet + (FDDI_HDR_LEN + SNAP_LEN);
	  }
      }
      break;

    default:
      offset = packet;
    }

  if (!offset)
    {
      not_ip++;
      return;
    }

  /* Offset is at IP hdr now: */
  iphdr = (struct ip*)offset;

  if (((ntohs(iphdr->ip_off) & IP_OFFMASK) != 0) ||
      ntohs(iphdr->ip_off) & IP_MF)
    fragmented++;

  switch (iphdr->ip_p)
    {
    case IPPROTO_UDP:
      t_udp++;
      break;
    case IPPROTO_TCP:
      t_tcp++;
      break;
    case IPPROTO_ICMP:
      t_icmp++;
      break;
    case IPPROTO_IGMP:
      t_igmp++;
      break;
    case IPPROTO_ESP:
      t_esp++;
      break;
    case IPPROTO_PIM:
      t_pim++;
      break;
    default:
      t_else++;
    }


  if (iphdr->ip_p != IPPROTO_UDP)
    {
      u_int16_t src_port = 0;
      u_int16_t dst_port = 0;

      if (iphdr->ip_p == IPPROTO_TCP)
	{
	  tcphdr = (struct tcphdr*)((u_char*)iphdr + 4*iphdr->ip_hl);
	  src_port = ntohs(tcphdr->th_sport);
	  dst_port = ntohs(tcphdr->th_dport);
	}

      if ((unsigned int)(ntohs(iphdr->ip_len) - 4*iphdr->ip_hl) > sizeof(struct udphdr))
	{
	  iphdr->ip_p = IPPROTO_UDP;

	  udphdr = (struct udphdr*)((u_char*)iphdr + 4*iphdr->ip_hl);
	  udphdr->uh_sport = htons(src_port);
	  udphdr->uh_dport = htons(dst_port);
	  udphdr->uh_ulen  = htons(ntohs(iphdr->ip_len) - 4*iphdr->ip_hl);
	  udphdr->uh_sum = udp_checksum(iphdr, udphdr);	  
	  changed++;


	  iphdr->ip_sum = ip_checksum(iphdr);
	}
      else
	{
	  too_small++;
	}
    }
  else
    {
      unchanged++;
    }

  return;
  pkthdr = NULL;
}


static void
handler(u_char *user, const struct pcap_pkthdr* pkthdr, const u_char * packet)
{
  adjust_header(pkthdr, (u_char*)packet);
  pcap_dump((u_char*)dumper, pkthdr, packet);

  return;
  user = NULL;
}


int
main(int argc, char **argv)
{
  if (argc < 3)
    {
      printf("USAGE: %s <input trace> <output trace>\n", argv[0]);
      exit(1);
    }

  input  = argv[1];
  output = argv[2];

  pcap = pcap_open_offline(input, errbuf);
  dumper = pcap_dump_open(pcap, output);
  pcap_loop(pcap, 0, handler, NULL);
  pcap_close(pcap);
  pcap_dump_close(dumper);

  printf("%i packets, %i changed, %i unchanged, %i too small, %i fragmented, %i not IP.\n",
	 total, changed, unchanged, too_small, fragmented, not_ip);

  printf("%i tcp, %i udp, %i icmp, %i igmp, %i esp, %i pim, %i other.\n",
	 t_tcp, t_udp, t_icmp, t_igmp, t_esp, t_pim, t_else);
  return (0);
}
